@@ -12,6 +12,7 @@ Features include:
  - comments shows table and column comments
  - compare ... to ... graphically compares results of two queries
  - commands are case-insensitive
+ - context-sensitive tab-completion for table names, column names, etc.
 
 Use 'help' within sqlpython for details.
 
@@ -139,7 +140,9 @@ UNION ALL
 SELECT 'DATABASE LINK' object_type, db_link, owner, 7 priority
 FROM   all_db_links dbl
 WHERE  dbl.db_link = :objName
-) ORDER BY priority ASC""",
+) ORDER BY priority ASC,
+           length(object_type) ASC,
+           object_type DESC""", # preference: PACKAGE before PACKAGE BODY, TABLE before INDEX
 'tabComments': """
 SELECT comments
 FROM    all_tab_comments
@@ -183,7 +186,7 @@ and c1.owner = :owner
 """
 }
 
-import sys, os, re, sqlpython, cx_Oracle, pyparsing
+import sys, os, re, sqlpython, cx_Oracle, pyparsing, re, completion
 from cmd2 import Cmd, make_option, options, Statekeeper
 
 if float(sys.version[:3]) < 2.3:
@@ -321,8 +324,8 @@ class Parser(object):
                         itms.append(sqlcommand[0])
         return itms
 
-bindScanner = Parser(pyparsing.Literal(':') + pyparsing.Word( pyparsing.alphanums + "_$#" ))
-
+bindScanner = Parser(pyparsing.Literal(':') + pyparsing.Word( pyparsing.alphanums + "_$#" ))   
+    
 def findBinds(target, existingBinds, givenBindVars = {}):
     result = givenBindVars
     for finding, startat, endat in bindScanner.scanner.scanString(target):
@@ -504,9 +507,58 @@ class sqlpyPlus(sqlpython.sqlpython):
 
     legalOracle = re.compile('[a-zA-Z_$#]')
 
+    def select_scalar_list(self, sql, binds={}):
+        self.curs.execute(sql, binds)
+        return [r[0] for r in self.curs.fetchall()]
+    
+    columnNameRegex = re.compile(
+        r'select\s+(.*)from',
+        re.IGNORECASE | re.DOTALL | re.MULTILINE)        
+    def completedefault(self, text, line, begidx, endidx):
+        segment = completion.whichSegment(line)
+        text = text.upper()
+        completions = []
+        if segment == 'select':
+            stmt = "SELECT column_name FROM user_tab_columns WHERE column_name LIKE '%s%%'"
+            completions = self.select_scalar_list(stmt % (text))
+            if not completions:
+                stmt = "SELECT column_name FROM all_tab_columns WHERE column_name LIKE '%s%%'"            
+                completions = self.select_scalar_list(stmt % (text))
+        if segment == 'from':
+            columnNames = self.columnNameRegex.search(line)
+            if columnNames:
+                columnNames = columnNames.group(1)
+                columnNames = [c.strip().upper() for c in columnNames.split(',')]
+                stmt1 = "SELECT table_name FROM all_tab_columns WHERE column_name = '%s' AND table_name LIKE '%s%%'"
+                for columnName in columnNames:
+                    # and if columnName is * ?
+                    completions.extend(self.select_scalar_list(stmt1 % (columnName, text)))                    
+        if segment in ('from', 'update', 'insert into') and (not completions):
+            stmt = "SELECT table_name FROM user_tables WHERE table_name LIKE '%s%%'"
+            completions = self.select_scalar_list(stmt % (text))
+            if not completions:
+                stmt = """SELECT table_name FROM user_tables WHERE table_name LIKE '%s%%'
+                      UNION
+                      SELECT DISTINCT owner FROM all_tables WHERE owner LIKE '%%%s'"""
+                completions = self.select_scalar_list(stmt % (text, text))
+        if segment in ('where', 'group by', 'order by', 'having', 'set'):
+            tableNames = completion.tableNamesFromFromClause(line)
+            if tableNames:
+                stmt = """SELECT column_name FROM all_tab_columns
+                          WHERE table_name IN (%s)""" % \
+                       (','.join("'%s'" % (t) for t in tableNames))
+            stmt = "%s AND column_name LIKE '%s%%'" % (stmt, text)
+            completions = self.select_scalar_list(stmt)
+        if not segment:
+            stmt = "SELECT object_name FROM all_objects WHERE object_name LIKE '%s%%'"
+            completions = self.select_scalar_list(stmt % (text))
+        return completions
+    
     rowlimitPattern = pyparsing.Word(pyparsing.nums)('rowlimit')
     terminatorPattern = (pyparsing.oneOf('; \\s \\S \\c \\C \\t \\x \\h')    
-                        ^ pyparsing.Literal('\n/\n')) ('terminator') + \
+                        ^ pyparsing.Literal('\n/') ^ \
+                        (pyparsing.Literal('\nEOF') + pyparsing.stringEnd)) \
+                        ('terminator') + \
                         pyparsing.Optional(rowlimitPattern)
     def do_select(self, arg, bindVarsIn=None, override_terminator=None):
         """Fetch rows from a table.
@@ -550,6 +602,7 @@ class sqlpyPlus(sqlpython.sqlpython):
             traceback.print_exc(file=sys.stdout)
         self.sqlBuffer.append(self.query)
 
+        
     @options([make_option('-f', '--full', action='store_true', help='get dependent objects as well')])
     def do_pull(self, arg, opts):
         """Displays source code."""
@@ -569,24 +622,29 @@ class sqlpyPlus(sqlpython.sqlpython):
                 except cx_Oracle.DatabaseError:
                     pass
 
-    @options([make_option('-i', '--insensitive', action='store_true', help='case-insensitive search'),
+    @options([make_option('-a','--all',action='store_true', help='Find in all schemas (not just my own)'),
+              make_option('-i', '--insensitive', action='store_true', help='case-insensitive search'),
               make_option('-c', '--col', action='store_true', help='find column'),
-              make_option('-t', '--table', action='store_true', help='find table')])                
+              make_option('-t', '--table', action='store_true', help='find table')])                    
     def do_find(self, arg, opts):
         """Finds argument in source code or (with -c) in column definitions."""
 
-        arg = self.parsed(arg).unterminated.upper()       
+        arg = self.parsed(arg).unterminated.upper()     
+        
         if opts.col:
-            self.do_select("owner, table_name, column_name from all_tab_columns where column_name like '%%%s%%'" % (arg))
+            sql = "owner, table_name, column_name from all_tab_columns where column_name like '%%%s%%'" % (arg)
         elif opts.table:
-            self.do_select("owner, table_name from all_tables where table_name like '%%%s%%'" % (arg))            
+            sql = "owner, table_name from all_tables where table_name like '%%%s%%'" % (arg)
         else:
             if opts.insensitive:
                 searchfor = "LOWER(text)"
                 arg = arg.lower()
             else:
                 searchfor = "text"
-            self.do_select("* from all_source where %s like '%%%s%%'" % (searchfor, arg))
+            sql = "* from all_source where %s like '%%%s%%'" % (searchfor, arg)
+        if not opts.all:
+            sql = '%s and owner = user' % (sql)
+        self.do_select(sql)
 
     @options([make_option('-a','--all',action='store_true',
                           help='Describe all objects (not just my own)')])
@@ -616,10 +674,9 @@ class sqlpyPlus(sqlpython.sqlpython):
             for q in descQ:
                 self.do_select(q,bindVarsIn={'object_name':object_name, 'owner':owner})
         elif object_type == 'PACKAGE':
-            self.curs.execute(descQueries['PackageObjects'][0], {'package_name':object_name, 'owner':owner})
-            packageContents = self.curs.fetchall()
-            for (packageObj_name,) in packageContents:
-                self.stdout.write(packageObj_name + '\n')
+            packageContents = self.select_scalar_list(descQueries['PackageObjects'][0], {'package_name':object_name, 'owner':owner})
+            for packageObj_name in packageContents:
+                self.stdout.write('Arguments to %s\n' % (packageObj_name))
                 self.do_select(descQueries['PackageObjArgs'][0],bindVarsIn={'package_name':object_name, 'owner':owner, 'object_name':packageObj_name})
     do_desc = do_describe
 
@@ -657,9 +714,9 @@ class sqlpyPlus(sqlpython.sqlpython):
         try:
             if len(parts) == 2:
                 owner, object_name = parts
-                self.curs.execute('SELECT object_type FROM all_objects WHERE owner = :owner AND object_name = :object_name',
-                                  {'owner': owner, 'object_name': object_name})
-                object_type = self.curs.fetchone()[0]
+                object_type = self.select_scalar_list('SELECT object_type FROM all_objects WHERE owner = :owner AND object_name = :object_name',
+                                  {'owner': owner, 'object_name': object_name}
+                                  )[0]
             elif len(parts) == 1:
                 object_name = parts[0]
                 self.curs.execute(queries['resolve'], {'objName':object_name})
@@ -915,6 +972,7 @@ class sqlpyPlus(sqlpython.sqlpython):
         self.stdout.write('\n'.join(result) + '\n')
 
     def do_cat(self, arg):
+        '''cat TABLENAME --> SELECT * FROM equivalent'''
         targets = arg.split()
         for target in targets:
             self.do_select('* from %s' % target)
@@ -922,29 +980,31 @@ class sqlpyPlus(sqlpython.sqlpython):
     @options([make_option('-i', '--ignore-case', dest='ignorecase', action='store_true', help='Case-insensitive search')])        
     def do_grep(self, arg, opts):
         """grep PATTERN TABLE - search for term in any of TABLE's fields"""    
-                
-        targets = []
-        for target in arg.split():
+
+        arg = self.parsed(arg)
+        targetnames = arg.unterminated.split()
+        pattern = targetnames.pop(0)
+        targets = [] 
+        for target in targetnames:
             if '*' in target:
-                self.curs.execute("SELECT owner, table_name FROM all_tables WHERE table_name LIKE '%s'" %
-                                  (target.upper().replace('*','%')))
+                self.curs.execute("SELECT owner, table_name FROM all_tables WHERE table_name LIKE '%s'%s" %
+                                  (target.upper().replace('*','%')), arg.terminator)
                 for row in self.curs:
                     targets.append('%s.%s' % row)
             else:
                 targets.append(target)
-        pattern = targets.pop(0)
         for target in targets:
             print target
             target = target.rstrip(';')
             sql = []
             try:
-                self.curs.execute('select * from %s where 1=0' % target)
+                self.curs.execute('select * from %s where 1=0' % target) # just to fill description
                 if opts.ignorecase:
                     sql = ' or '.join("LOWER(%s) LIKE '%%%s%%'" % (d[0], pattern.lower()) for d in self.curs.description)                                        
                 else:
                     sql = ' or '.join("%s LIKE '%%%s%%'" % (d[0], pattern) for d in self.curs.description)
                 sql = '* FROM %s WHERE %s' % (target, sql)
-                self.do_select(sql)
+                self.do_select('%s%s%s' % (sql, arg.terminator, arg.rowlimit))
             except Exception, e:
                 print e
                 import traceback
@@ -955,7 +1015,7 @@ class sqlpyPlus(sqlpython.sqlpython):
         object_type, owner, object_name = self.resolve(arg)
         if object_type == 'TABLE':
             self.do_select(queries['refs'],bindVarsIn={'object_name':object_name, 'owner':owner})
-
+    
 def _test():
     import doctest
     doctest.testmod()
